// Copyright Krzysztof Czainski 2011
// Distributed under the Boost Software License, Version 1.0.
//    (See accompanying file LICENSE_1_0.txt or copy at
//          http://www.boost.org/LICENSE_1_0.txt)

/*
 * @file main.cpp
 *
 * @since 14-07-2012
 * @author Krzysztof Czainski
 */

#include "ptr.hpp"

#include <boost/format.hpp>

using namespace std;

int ptr_s = 0;

struct A
{
    static int a, c;
    A() { ++a; ++c; cout << "_A "; }
    A( A const& ) { ++a; ++c; cout << "cA "; }
    virtual ~A() { ++a; --c; cout << "~A "; }
};
int A::a = 0;
int A::c = 0;
struct B : A
{
    static int b, d;
    B() { ++b; ++d; cout << "_B "; }
    B( B const& x ) : A(x) { ++b; ++d; cout << "cB "; }
    ~B() { ++b; --d; cout << "~B "; }
};
int B::b = 0;
int B::d = 0;

ptr<A> make_a()
{
    return ptr<A>( new A );
}

ptr<B> make_b()
{
    return ptr<B>( new B );
}

void trace( char const* msg )
{
    cout << endl << boost::format("%34s P:%d A:%d B:%d eA:%d eB:%d") % msg %
            ptr_s % A::a % B::b % A::c % B::d << endl;
    ptr_s = 0; A::a = 0; B::b = 0;
}

#define TRACE( x ) x; trace( #x );

int main()
{
    {
        TRACE( ptr<A> z )
        TRACE( ptr<A> a( new A ) )
        TRACE( ptr<B> b( new B ) )
        TRACE( ptr<A> c( new B ) )
        TRACE( ptr<A> d( a ) )
        TRACE( ptr<A> e( b ) )
        TRACE( ptr<A> f( boost::move(a) ) )
        TRACE( ptr<A> g( boost::move(b) ) )
        TRACE( ptr<B> h( new B ) )
        TRACE( ptr<A> i( boost::move(c) ) )
        TRACE( ptr<A> j( i ) ) // slice

        TRACE( ptr<A> x( new A ) )
        TRACE( x = f )
        TRACE( x = h )
        TRACE( x = boost::move(f) )
        TRACE( x = boost::move(g) )
        TRACE( x = boost::move(h) )
        TRACE( x = i ) // slice
    }
    trace( "" );
    {
        TRACE( ptr<A> a = make_a() )
        TRACE( ptr<A> b = make_b() )
        TRACE( ptr<A> c( make_b() ) )
        TRACE( a = make_a() )
        TRACE( b = make_b() )
    }
    trace( "" );
}
